from dash import Dash, html, dcc, Output, Input
import dash_ag_grid as dag

from . import utils


def test_cc001_custom_comparator(dash_duo):
    import dash_ag_grid as dag
    from dash import Dash, html, dcc
    import pandas as pd

    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    # basic columns definition with column defaults
    columnDefs = [
        {"field": "athlete", "sort": "desc"},
        {"field": "age", "width": 90},
        {"field": "country"},
        {"field": "year", "width": 90, "unSortIcon": True},
        {"field": "date", "comparator": {"function": "dateComparator"}},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]
    defaultColDef = {
        "width": 170,
        "sortable": True,
    }

    app.layout = html.Div(
        [
            dcc.Markdown("Date Sort Comparator Example"),
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records")[:100],
                defaultColDef=defaultColDef,
                id="grid",
            ),
        ],
        style={"margin": 20},
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "Yannick Agnel")

    grid.get_header_cell(4).click()
    grid.wait_for_cell_text(0, 0, "Aleksey Nemov")
    grid.wait_for_cell_text(0, 4, "1/10/2000")

    grid.get_header_cell(4).click()
    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    grid.wait_for_cell_text(0, 4, "12/8/2012")



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output
import pandas as pd



def test_pi001_pivot_column_order(dash_duo):
    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    columnDefs = [
        {"field": "country", "rowGroup": True},
        {"field": "sport", "pivot": True, "pivotComparator": {"function": "sortColumns"}},
        {"field": "year"},
        {"field": "date"},
        {"field": "gold", "aggFunc": "sum"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                dashGridOptions={
                    "autoGroupColumnDef": {"minWidth": 250},
                    "pivotMode": True,
                },
                enableEnterpriseModules=True,
            ),
            html.Div(id="col-order"),
        ],
    )

    @app.callback(
        Output("col-order", "children"),
        Input("grid", "columnState"),
        prevent_initial_call=True,
    )
    def display_state(col_state):

        # Pivot columns are identified in `columnState` by `colId`.
        # If not specified by the user, the `colId` is auto-generated with a `pivot_{field name}` prefix.

        pivot_columns = [
            c["colId"] for c in col_state if c["colId"].startswith("pivot_sport")
        ]

        # By default, the columns are sorted in ascending order.
        # This app uses a custom `pivotComparator` function to sort the columns in descending order.
        # The following check verifies if the columns are correctly sorted in descending order.
        descending = pivot_columns == sorted(pivot_columns, reverse=True)
        return f"pivot columns sorted in reverse order? {descending}"

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal(
        "#col-order", "pivot columns sorted in reverse order? True"
    )





import dash
import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output, State, ctx
import json

from . import utils


def test_sp001_selection_persistence(dash_duo):
    app = Dash(__name__)

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    columnDefs = [
        {
            "headerName": "id",
            "valueGetter": {"function": "params.node.id"},
            "checkboxSelection": True,
        },
        {"field": "make"},
        {"field": "model"},
        {
            "field": "price",
            "cellRenderer": "agAnimateShowChangeCellRenderer",
        },
    ]

    app.layout = html.Div(
        [
            html.Button("Update selected", id="transactions-update"),
            html.Button("Remove Selected", id="transactions-remove"),
            html.Button("Add Rows", id="transactions-add"),
            html.Button("Clear", id="transactions-clear"),
            html.Button("Start Over", id="transactions-start"),
            html.Button(id="ids"),
            html.Button(id="function"),
            html.Button(id="rowInfo"),
            dag.AgGrid(
                id="grid",
                rowData=rowData,
                columnDefs=columnDefs,
                defaultColDef={"flex": 1},
                dashGridOptions={"rowSelection": "multiple"},
                getRowId="params.data.make",
            ),
            html.Div(id="selectedRows"),
        ],
    )

    @app.callback(
        Output("grid", "rowData"),
        Input("transactions-clear", "n_clicks"),
        Input("transactions-start", "n_clicks"),
    )
    def update_rowdata(*_):
        if ctx.triggered_id == "transactions-clear":
            return []
        return rowData

    @app.callback(
        Output("grid", "rowTransaction"),
        Input("transactions-remove", "n_clicks"),
        Input("transactions-update", "n_clicks"),
        Input("transactions-add", "n_clicks"),
        State("grid", "selectedRows"),
    )
    def update_rowdata(n1, n2, n3, selection):
        if ctx.triggered_id == "transactions-remove":
            if selection is None:
                return dash.no_update
            return {"remove": selection}

        if ctx.triggered_id == "transactions-update":
            if selection is None:
                return dash.no_update
            for row in selection:
                row["price"] = row["price"] + n2
            return {"update": selection}

        if ctx.triggered_id == "transactions-add":
            newRows = [row.copy() for row in rowData]
            for row in newRows:
                row["make"] = row["make"] + str(n3)
            return {"add": newRows}

    @app.callback(Output("selectedRows", "children"), Input("grid", "selectedRows"))
    def selectedRows(s):
        return json.dumps(s)

    @app.callback(
        Output("grid", "selectedRows"),
        Input("ids", "n_clicks"),
        Input("function", "n_clicks"),
        Input("rowInfo", "n_clicks"),
    )
    def setSelections(n, n1, n2):
        if n or n1 or n2:
            if ctx.triggered_id == "ids":
                return {"ids": ["Toyota"]}
            if ctx.triggered_id == "function":
                return {"function": "params.data.make === 'Porsche'"}
            return [{"make": "Ford", "model": "Mondeo", "price": 32000}]
        return []

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "Toyota")
    grid.get_cell(1, 0).click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Ford", "model": "Mondeo", "price": 32000}]'
    )

    dash_duo.find_element("#transactions-update").click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Ford", "model": "Mondeo", "price": 32001}]'
    )

    dash_duo.find_element("#transactions-start").click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Ford", "model": "Mondeo", "price": 32000}]'
    )

    dash_duo.find_element("#ids").click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Toyota", "model": "Celica", "price": 35000}]'
    )

    dash_duo.find_element("#function").click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Porsche", "model": "Boxster", "price": 72000}]'
    )

    dash_duo.find_element("#rowInfo").click()
    dash_duo.wait_for_text_to_equal(
        "#selectedRows", '[{"make": "Ford", "model": "Mondeo", "price": 32000}]'
    )



import dash_ag_grid as dag
from dash import Dash, html, dcc
from . import utils
from dash.testing.wait import until
import pandas as pd
import time


def test_ca001_cell_change_anitmation(dash_duo):
    app = Dash(__name__)

    df = pd.DataFrame(data=[[1, 2, 3], [2, 3, 4], [9, 8, 6]], columns=list("abc"))
    df.reset_index(inplace=True)

    columnDefs = [
        {
            "headerName": "Editable A",
            "field": "a",
            "editable": True,
            "resizable": True,
        },
        {
            "headerName": "Editable B",
            "field": "b",
            "editable": True,
        },
        {
            "headerName": "API C",
            "field": "c",
            "cellRenderer": "agAnimateShowChangeCellRenderer",
        },
        {
            "headerName": "Total",
            "valueGetter": {
                "function": "Number(params.data.a) + Number(params.data.b) + Number(params.data.c)"
            },
            "cellRenderer": "agAnimateShowChangeCellRenderer",
        },
        {
            "headerName": "Average",
            "valueGetter": {
                "function": "Math.round((Number(params.data.a) + Number(params.data.b) + Number(params.data.c)) * 10 / 3) /10"
            },
            "cellRenderer": "agAnimateShowChangeCellRenderer",
        },
    ]

    defaultColDef = {
        "type": ["numberColumn", "rightAligned"],
        "filter": "agNumberColumnFilter",
        "resizable": True,
        "sortable": True,
    }

    app.layout = html.Div(
        [
            dcc.Markdown(
                "This grid demonstrates the renderer for animating changes.  Try entering new values into the editable columns and press the button to update column c in a callback"
            ),
            html.Button("Update column C", id="live-data-grid-btn"),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef=defaultColDef,
                # setting a row ID is required when updating data in a callback
                getRowId="params.data.index",
            ),
        ],
        style={"margin": 20},
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "1")

    ### testing animations
    grid.get_cell(1, 1).send_keys("50")
    grid.get_cell(1, 2).click()
    start = time.time()
    until(
        lambda: "47" in grid.get_cell_animation(1, 3).get_attribute("innerText"),
        timeout=3,
    )
    until(
        lambda: "47" not in grid.get_cell_animation(1, 3).get_attribute("innerText"),
        timeout=3,
    )
    end = time.time()
    assert (end - start) > 1.8

    grid.get_cell(2, 1).send_keys("50")
    grid.get_cell(2, 2).click()
    start = time.time()
    until(
        lambda: "42" in grid.get_cell_animation(2, 3).get_attribute("innerText"),
        timeout=3,
    )
    until(
        lambda: "42" not in grid.get_cell_animation(2, 3).get_attribute("innerText"),
        timeout=3,
    )
    end = time.time()
    assert (end - start) > 1.8



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output, no_update
from . import utils
import time
from dash.testing.wait import until
import json


def test_td001_tree_data(dash_duo):
    app = Dash(__name__)

    rowData = [
        {
            "orgHierarchy": ["Erica Rogers"],
            "jobTitle": "CEO",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett"],
            "jobTitle": "Exec. Vice President",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett", "Esther Baker"],
            "jobTitle": "Director of Operations",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
            ],
            "jobTitle": "Fleet Coordinator",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
                "Leah Flowers",
            ],
            "jobTitle": "Parts Technician",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
                "Tammy Sutton",
            ],
            "jobTitle": "Service Technician",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Derek Paul",
            ],
            "jobTitle": "Inventory Control",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett", "Francis Strickland"],
            "jobTitle": "VP Sales",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Morris Hanson",
            ],
            "jobTitle": "Sales Manager",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Todd Tyler",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Bennie Wise",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Joel Cooper",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Permanent",
        },
    ]

    grid = html.Div(
        [
            dag.AgGrid(
                columnDefs=[
                    # we're using the auto group column by default!
                    {"field": "jobTitle"},
                    {"field": "employmentType"},
                ],
                defaultColDef={
                    "flex": 1,
                },
                dashGridOptions={
                    "autoGroupColumnDef": {
                        "headerName": "Organisation Hierarchy",
                        "minWidth": 300,
                        "cellRendererParams": {
                            "suppressCount": True,
                        },
                    },
                    "groupDefaultExpanded": -1,
                    "getDataPath": {"function": "getDataPath(params)"},
                    "treeData": True,
                },
                rowData=rowData,
                enableEnterpriseModules=True,
                id="grid",
            ),
            html.Hr(),
        ]
    )

    app.layout = html.Div(
        [
            grid,
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Erica Rogers")

    ### testing tree functions
    grid.wait_for_cell_text(1, 0, "Malcolm Barrett")
    grid.wait_for_cell_text(3, 0, "Brittany Hanson")
    grid.get_cell_collapsable(2, 0).click()
    grid.wait_for_cell_text(3, 0, "Francis Strickland")
    grid.get_cell_expandable(2, 0).click()
    grid.wait_for_cell_text(3, 0, "Brittany Hanson")


def test_td002_tree_data(dash_duo):
    app = Dash(__name__)

    rowData = [
        {
            "orgHierarchy": ["Erica Rogers"],
            "jobTitle": "CEO",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett"],
            "jobTitle": "Exec. Vice President",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett", "Esther Baker"],
            "jobTitle": "Director of Operations",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
            ],
            "jobTitle": "Fleet Coordinator",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
                "Leah Flowers",
            ],
            "jobTitle": "Parts Technician",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Brittany Hanson",
                "Tammy Sutton",
            ],
            "jobTitle": "Service Technician",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Esther Baker",
                "Derek Paul",
            ],
            "jobTitle": "Inventory Control",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": ["Erica Rogers", "Malcolm Barrett", "Francis Strickland"],
            "jobTitle": "VP Sales",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Morris Hanson",
            ],
            "jobTitle": "Sales Manager",
            "employmentType": "Permanent",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Todd Tyler",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Bennie Wise",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Contract",
        },
        {
            "orgHierarchy": [
                "Erica Rogers",
                "Malcolm Barrett",
                "Francis Strickland",
                "Joel Cooper",
            ],
            "jobTitle": "Sales Executive",
            "employmentType": "Permanent",
        },
    ]

    grid = html.Div(
        [
            dag.AgGrid(
                columnDefs=[
                    # we're using the auto group column by default!
                    {"field": "jobTitle"},
                    {"field": "employmentType"},
                ],
                defaultColDef={
                    "flex": 1,
                },
                dashGridOptions={
                    "autoGroupColumnDef": {
                        "headerName": "Organisation Hierarchy",
                        "minWidth": 300,
                        "cellRendererParams": {
                            "suppressCount": True,
                        },
                    },
                    "getDataPath": {"function": "getDataPath(params)"},
                    "treeData": True,
                },
                rowData=rowData,
                enableEnterpriseModules=True,
                id="grid",
            ),
            html.Hr(),
            html.Button(id="reset"),
        ]
    )

    @app.callback(Output("grid", "rowData"), Input("reset", "n_clicks"))
    def resetRowData(n):
        if n:
            return rowData
        return no_update

    app.layout = html.Div(
        [
            grid,
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Erica Rogers")

    ### testing tree functions
    grid.get_cell_expandable(0, 0).click()
    grid.wait_for_cell_text(1, 0, "Malcolm Barrett")
    grid.get_cell_expandable(1, 0).click()
    grid.wait_for_cell_text(2, 0, "Esther Baker")
    grid.get_cell_expandable(2, 0).click()
    grid.wait_for_cell_text(3, 0, "Brittany Hanson")
    dash_duo.find_element("#reset").click()

    grid.wait_for_cell_text(3, 0, "Brittany Hanson")



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output, State, Patch
import plotly.express as px
from . import utils
import json
from selenium.webdriver.common.by import By
from dash.testing.wait import until
import time


df = px.data.medals_wide()

columnDefs = []
for i in df.columns:
    if i == "nation":
        columnDefs.append({"field": i, "editable": False})
    else:
        columnDefs.append({"field": i})


def test_el001_event_listener(dash_duo):
    app = Dash(__name__, suppress_callback_exceptions=True)
    app.layout = html.Div(
        [
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef={"editable": True},
                id="grid",
                getRowId="params.data.nation",
                eventListeners={'cellContextMenu': ['showOutput(params, setGridProps)']},
                dashGridOptions={'suppressContextMenu': True, "preventDefaultOnContextMenu": True}
            ),
            html.Div(id="log")
        ]
    )

    app.clientside_callback(
        """function countEvents(cellClicked) {
            return JSON.stringify(cellClicked)
        }""",
        Output("log", "children"),
        Input("grid", "cellClicked"),
        prevent_initial_call=True,
        suppress_callback_exception=True
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "South Korea")

    # Test left click.
    grid.get_cell(1, 2).click()
    until(lambda: json.loads(dash_duo.find_element('#log').text or "{}").get('value') == 15, timeout=3)

    # Test right click
    action = utils.ActionChains(dash_duo.driver)
    action.context_click(grid.get_cell(0, 2)).perform()
    until(lambda: json.loads(dash_duo.find_element('#log').text).get('value') == 13, timeout=3)
    until(lambda: json.loads(dash_duo.find_element('#log').text).get('contextMenu'), timeout=3)

def test_el002_event_listener(dash_duo):
    app = Dash(__name__, suppress_callback_exceptions=True)
    app.layout = html.Div(
        [
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef={"editable": True},
                id="grid",
                getRowId="params.data.nation",
                eventListeners={'cellClicked': ['dash_clientside.set_props("log", {children: "rawr"})']},
            ),
            html.Div(id="log")
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "South Korea")

    # Test left click.
    grid.get_cell(1, 2).click()
    until(lambda: dash_duo.find_element('#log').text == "rawr", timeout=3)

def test_el003_event_listener(dash_duo):
    app = Dash(__name__, suppress_callback_exceptions=True)
    app.layout = html.Div([
        dag.AgGrid(id='grid',
                   columnDefs=[{'field': 'test'}],
                   rowData=[{'test': '1'}],
                   eventListeners={'cellClicked': ['TestEvent(params, setEventData)']}
                   ),
        html.Div(id='output', children=[])
    ]
    )

    @app.callback(
        Output('output', 'children'),
        Input('grid', 'eventData')
    )
    def show_event_data(v):
        children = Patch()
        if v:
            children.append(html.Div(json.dumps(v)))
        return children

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "1")

    for i in range(5):
        grid.get_cell(0, 0).click()

        # Assert that the output element has children
        until(lambda: len(dash_duo.find_element("#output").find_elements(By.XPATH, "*")) == (i + 1), timeout=3)





from dash import Dash, html, dcc, Output, Input
import dash_ag_grid as dag

from . import utils


def test_vr001_virtual_row_data(dash_duo):
    app = Dash(__name__)

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    columnDefs = [
        {"headerName": "Make", "field": "make", "sortable": True},
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown(
                "Try filtering the data in the grid using the inline filters (click the hamburger menu in each column). Though the data is still in `rowData`, you can view the virtual row data in callbacks by watching the `virtualRowData` property."
            ),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
            ),
            html.Div(id="data-after-filter"),
        ]
    )

    @app.callback(
        Output("data-after-filter", "children"),
        Input("grid", "virtualRowData"),
    )
    def get_virtual_data(virtual):
        return str(virtual)

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter",
        "[{'make': 'Toyota', 'model': 'Celica', 'price': 35000}, "
        "{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}, "
        "{'make': 'Porsche', 'model': 'Boxster', 'price': 72000}]",
    )

    grid.set_filter(0, "F")

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter", "[{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}]"
    )

    grid.set_filter(0, "")

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter",
        "[{'make': 'Toyota', 'model': 'Celica', 'price': 35000}, "
        "{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}, "
        "{'make': 'Porsche', 'model': 'Boxster', 'price': 72000}]",
    )

    grid.get_header_cell(0).click()

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter",
        "[{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}, "
        "{'make': 'Porsche', 'model': 'Boxster', 'price': 72000},"
        " {'make': 'Toyota', 'model': 'Celica', 'price': 35000}]",
    )

    grid.get_header_cell(0).click()

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter",
        "[{'make': 'Toyota', 'model': 'Celica', 'price': 35000}, "
        "{'make': 'Porsche', 'model': 'Boxster', 'price': 72000}, "
        "{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}]",
    )

    grid.get_header_cell(0).click()

    dash_duo.wait_for_text_to_equal(
        "#data-after-filter",
        "[{'make': 'Toyota', 'model': 'Celica', 'price': 35000}, "
        "{'make': 'Ford', 'model': 'Mondeo', 'price': 32000}, "
        "{'make': 'Porsche', 'model': 'Boxster', 'price': 72000}]",
    )



"""
Working with raw html in Markdown component
"""
import dash_ag_grid as dag
import dash
from dash import html, dcc
from . import utils
from dash.testing.wait import until


def test_mc001_markdown_components(dash_duo):
    app = dash.Dash(__name__)

    columnDefs = [
        {
            "headerName": "Make",
            "field": "make",
            "sortable": True,
            "cellRenderer": "markdown",
        },
        {"headerName": "Model", "field": "model", "cellRenderer": "markdown"},
        {
            "headerName": "Link",
            "field": "link",
            "cellRenderer": "markdown",
            "linkTarget": "_blank",
        },
        {"headerName": "Image", "field": "image", "cellRenderer": "markdown"},
    ]

    columnDefs_allow_html = [
        {
            "headerName": "Make",
            "field": "make",
            "sortable": True,
            "cellRenderer": "markdown",
        },
        {
            "headerName": "Model",
            "field": "model",
            "cellRenderer": "markdown",
        },
        {
            "headerName": "Link",
            "field": "link",
            "cellRenderer": "markdown",
            "linkTarget": "_self",
        },
        {"headerName": "Image", "field": "image", "cellRenderer": "markdown"},
    ]

    rowData = [
        {
            "make": "*Toyota* in italics",
            "model": "`code snippet`",
            "link": "**[Bold link](#)**",
            "image": "{0} {0} {0} {0} {0}".format(
                "![alt text: rain](https://www.ag-grid.com/example-assets/weather/rain.png)"
            ),
        },
        {
            "make": "**Ford** in bold",
            "model": "Mondeo",
            "link": '<a href="#" target="_blank">Link to new tab</a>',
            "image": "{0} {0} {0} {0}".format(
                "![alt text: sun](https://www.ag-grid.com/example-assets/weather/sun.png)"
            ),
        },
        {
            "make": "***Porsche*** in both",
            "model": "<b>Boxster</b> in HTML bold",
            "link": "[Example](#)",
            "image": "![alt text: rain](https://www.ag-grid.com/example-assets/weather/rain.png)",
        },
    ]

    rowData2 = [{k: row[k] for k in row if k != "link"} for row in rowData]

    raw_html_example1 = html.Div(
        [
            dcc.Markdown(
                "This grid has both Markdown and raw HTML. By default, raw HTML is not rendered."
            ),
            dag.AgGrid(
                id="not_dangerous",
                columnSize="sizeToFit",
                columnDefs=columnDefs,
                rowData=rowData,
            ),
            html.Hr(),
        ]
    )

    raw_html_example2 = html.Div(
        [
            dcc.Markdown(
                "This grid has both Markdown and raw HTML. `dangerously_allow_code=True` so the links render"
            ),
            dag.AgGrid(
                id="dangerous",
                columnSize="sizeToFit",
                columnDefs=columnDefs_allow_html,
                rowData=rowData,
                dangerously_allow_code=True,
            ),
        ]
    )

    raw_html_example3 = html.Div(
        [
            dcc.Markdown(
                "This grid has both Markdown and raw HTML. By default, raw HTML is not rendered."
            ),
            dag.AgGrid(
                id="not_dangerous2",
                columnSize="sizeToFit",
                columnDefs=columnDefs,
                rowData=rowData2,
            ),
            html.Hr(),
        ]
    )

    app.layout = html.Div(
        [raw_html_example1, raw_html_example2, raw_html_example3],
        style={"flexWrap": "wrap"},
    )

    dash_duo.start_server(app)

    safe_grid = utils.Grid(dash_duo, "not_dangerous")
    dangerous_grid = utils.Grid(dash_duo, "dangerous")
    safe2_grid = utils.Grid(dash_duo, "not_dangerous2")

    for grid in [safe_grid, dangerous_grid]:
        grid.wait_for_cell_text(0, 0, "Toyota in italics")
        assert (
            grid.get_cell(0, 0).get_attribute("innerHTML")
            == '<div class="agGrid-Markdown"><div><em>Toyota</em> in italics</div></div>'
        )
        grid.wait_for_cell_text(1, 1, "Mondeo")
        grid.wait_for_cell_text(2, 2, "Example")
        assert (
            grid.get_cell(2, 3).get_attribute("innerHTML")
            == '<div class="agGrid-Markdown">'
            '<div><img src="https://www.ag-g'
            'rid.com/example-assets/weather/rain.png" '
            'alt="alt text: rain"></div></div>'
        )

    assert (
        safe_grid.get_cell(2, 2).get_attribute("innerHTML")
        == '<div class="agGrid-Markdown"><div><a target="_blank" href="#" rel="noopener noreferrer nofollow">Example</a></div></div>'
    )
    safe_grid.wait_for_cell_text(
        1, 2, '<a href="#" target="_blank">Link to new tab</a>'
    )

    assert (
        dangerous_grid.get_cell(2, 2).get_attribute("innerHTML")
        == '<div class="agGrid-Markdown"><div><a target="_self" href="#">Example</a></div></div>'
    )
    dangerous_grid.wait_for_cell_text(1, 2, "Link to new tab")
    assert (
        dangerous_grid.get_cell(1, 2).get_attribute("innerHTML")
        == '<div class="agGrid-Markdown"><div><a target="_blank" href="#" rel="noopener noreferrer nofollow">Link to new tab</a></div></div>'
    )

    assert safe2_grid.get_cell(0, 2).text == ""
    assert safe2_grid.get_cell(1, 2).text == ""
    assert safe2_grid.get_cell(2, 2).text == ""



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output, State
import plotly.express as px
from . import utils

df = px.data.medals_wide()

columnDefs = []
for i in df.columns:
    if i == "nation":
        columnDefs.append({"field": i, "editable": False})
    else:
        columnDefs.append({"field": i})


def test_cv001_cell_value_changed(dash_duo):
    app = Dash(__name__)

    app.layout = html.Div(
        [
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef={"editable": True},
                id="information",
                getRowId="params.data.nation",
            ),
            dag.AgGrid(
                id="history",
                columnDefs=[{"field": "Key", "checkboxSelection": True}]
                           + [{"field": i} for i in ["Column", "OldValue", "NewValue"]],
                rowData=[],
                dashGridOptions={
                    "rowSelection": {
                        "mode": 'singleRow'
                    },
                }
            ),
        ],
        style={"margin": 20},
    )

    app.clientside_callback(
        """function (n){if (n) {return true} return window.dash_clientside.no_update}""",
        Output("show_history", "is_open"),
        Input("viewHistory", "n_clicks"),
    )

    app.clientside_callback(
        """function addToHistory(changes) {
            if (changes) {
                newData = []
                for (let i = 0; i < changes.length; i++) {
                    data = changes[i];
                    reloadData = {...data.data};
                    reloadData[data.colId] = data.oldValue;
                    newData.push({Key: data.rowId, Column: data.colId, OldValue: data.oldValue,
                    NewValue: data.value, reloadData});
                }
                return {'add': newData}
            }
            return window.dash_clientside.no_update
        }""",
        Output("history", "rowTransaction"),
        Input("information", "cellValueChanged"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        """function reloadHistory(data) {
            if (data.length) {
                return [{'update': [data[0].reloadData], 'async': false}, {'remove': [data[0]], 'async': false}]
            }
            return [null]*2
        }""",
        Output("information", "rowTransaction"),
        Output("history", "rowTransaction", allow_duplicate=True),
        Input("history", "selectedRows"),
        prevent_initial_call=True,
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "information")
    hist = utils.Grid(dash_duo, "history")

    grid.wait_for_cell_text(0, 0, "South Korea")

    ### testing history
    grid.get_cell(0, 1).send_keys("50")
    grid.get_cell(1, 2).click()

    hist.wait_for_rendered_rows(1)

    hist.element_click_cell_checkbox(0, 0)
    hist.wait_for_rendered_rows(0)

    grid.get_cell(0, 1).send_keys("50")
    grid.get_cell(1, 2).click()

    hist.wait_for_rendered_rows(1)

    ## twice for good measure
    hist.element_click_cell_checkbox(0, 0)
    hist.wait_for_rendered_rows(0)

    grid.get_cell(0, 1).send_keys("50")
    grid.get_cell(1, 2).click()

    hist.wait_for_rendered_rows(1)


def test_cv001_cell_value_changed_multi(dash_duo):
    app = Dash(__name__)
    app.layout = html.Div(
        [
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef={"editable": True},
                id="grid",
                getRowId="params.data.nation",
                dashGridOptions={'editType':'fullRow'}
            ),
            html.Div(id="log")
        ]
    )

    app.clientside_callback(
        """function countEvents(changes) {
            console.log("FIRE");
            return changes? changes.length : 0;
        }""",
        Output("log", "children"),
        Input("grid", "cellValueChanged"),
        prevent_initial_call=True,
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "South Korea")

    # Test single event.
    grid.get_cell(0, 1).click()
    grid.get_cell(0, 1).send_keys("50")
    grid.get_cell(1, 2).click()
    dash_duo.wait_for_text_to_equal('#log', "1")

    # Test multi event.
    grid.get_cell(0, 1).send_keys("20")
    grid.get_cell_editing_input(0, 2).send_keys("20")
    grid.get_cell(0, 2).click()
    grid.get_cell(1, 2).click()
    dash_duo.wait_for_text_to_equal('#log', "2")



from selenium.webdriver import Keys

import dash_ag_grid as dag
from dash import Dash, html
from . import utils

def test_cd001_cell_data_types_override(enforced_locale, dash_duo):
    app = Dash(__name__)

    rowData = [
        {"weight": 0.074657, "date": "01/01/2024"},
        {"weight": 0.06948567, "date": "02/01/2024"},
        {"weight": 0.02730574, "date": "03/01/2024"},
        {"weight": 0.0182345, "date": "04/01/2024"},
    ]

    columnDefs = [
        {"field": "weight", "cellDataType": "percentage"},
        {"field": "date", "cellDataType": "dateString"},
    ]

    # Only for second grid
    dataTypeDefinitions = {
        "percentage": {
            "baseDataType": "number",
            "extendsDataType": "number",
            "valueFormatter": {
                "function": "params.value == null ? '' :  d3.format(',.1%')(params.value)"
            },
        },
        "dateString": {
            "baseDataType": 'dateString',
            "extendsDataType": 'dateString',
            "valueParser": {
                "function": r"params.newValue != null && !!params.newValue.match(/\d{2}\/\d{2}\/\d{4}/) ? params.newValue : null"
            },
            "valueFormatter": {"function": "params.value == null ? '' : params.value"},
            "dataTypeMatcher": {"function": r"params != null && !!params.match(/\d{2}\/\d{2}\/\d{4}/)"},
            "dateParser": {"function": "dateParser(params)"},
            "dateFormatter": {"function": "dateFormatter(params)"},
        },
    }
    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid-cell-data-types-override-full-JS",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={"filter": True, "editable": True},
                dashGridOptions={"dataTypeDefinitions": {"function": "dataTypeDefinitions"}},
            ),
            dag.AgGrid(
                id="grid-cell-data-types-override",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={"filter": True, "editable": True},
                dashGridOptions={"dataTypeDefinitions": dataTypeDefinitions},
            ),
        ],
    )

    dash_duo.start_server(app)

    action = utils.ActionChains(dash_duo.driver)

    # same tests for both grids
    for id in ["grid-cell-data-types-override-full-JS", "grid-cell-data-types-override"]:
        grid = utils.Grid(dash_duo, id)

        # test overriden number cell data type
        action.double_click(grid.get_cell(0, 0)).perform()
        date_input_element = dash_duo.find_element(f'#{grid.id} .ag-number-field-input')
        date_input_element.send_keys("0.1" + Keys.ENTER)

        grid.wait_for_cell_text(0, 0, "10.0%")

        # test overriden dateString cell data type
        action.double_click(grid.get_cell(0, 1)).perform()
        date_input_element = dash_duo.find_element(f'#{grid.id} .ag-date-field-input')
        date_input_element.send_keys("01172024" + Keys.ENTER)

        grid.wait_for_cell_text(0, 1, "17/01/2024")


def test_cd002_column_types_formatting(dash_duo):
    app = Dash(__name__)

    rowData = [
        {"col1": 0.12345, "col2": 0.98765},
        {"col1": 0.5, "col2": 0.25},
    ]

    columnDefs = [
        {
            "field": "col1",
            "type": "rightAligned",
            "valueFormatter": {"function": "d3.format('.3f')(params.value)"},
        },
        {
            "field": "col2",
            "type": "float",
        },
    ]

    dashGridOptions = {
        "columnTypes": {
            "float": {
                "cellClass": "ag-right-aligned-cell",
                "headerClass": "ag-right-aligned-header",
                "valueFormatter": {
                    "function": "d3.format('.3f')(params.value)"
                },
            }
        }
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid-column-types",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={"editable": True},
                dashGridOptions=dashGridOptions,
            )
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid-column-types")

    action = utils.ActionChains(dash_duo.driver)

    # ---- Test col1 formatter ----
    # edit valye
    action.double_click(grid.get_cell(0, 0)).perform()
    input_el = dash_duo.find_element("#grid-column-types .ag-input-field-input")
    input_el.send_keys("0.1" + Keys.ENTER)

    # expect formatted to 3 decimals
    grid.wait_for_cell_text(0, 0, "0.100")

    # ---- Test col2 formatter via column type ----
    action.double_click(grid.get_cell(0, 1)).perform()
    input_el = dash_duo.find_element("#grid-column-types .ag-input-field-input")
    input_el.send_keys("0.2" + Keys.ENTER)

    grid.wait_for_cell_text(0, 1, "0.200")


import dash_ag_grid as dag
from dash import Dash, html, dcc
from . import utils


def test_cd001_cell_renderer_function(dash_duo):
    app = Dash(__name__)

    rowData = [
        {"size": 0, "is_available": True},
        {"size": 1, "is_available": False},
        {"size": 2, "is_available": True},
    ]

    columnDefs = [
        {
            "field": "size",
            "cellRenderer": {"function": "params.value < 1 ? 'small' : params.value < 2 ? 'medium' : 'large'"},
        },
        {
            "field": "is_available",
            "cellRenderer": {"function": "params.value ? 'yes' : 'no'"},
        },
    ]

    app.layout = html.Div(
        [
            dcc.Markdown(
                "This grid uses a javascript function to display computed values for each cell, rather than the raw numbers."
            ),
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=rowData,
                id="grid",
            ),
        ],
        style={"margin": 20},
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "small")
    grid.wait_for_cell_text(0, 1, "yes")
    grid.wait_for_cell_text(1, 0, "medium")
    grid.wait_for_cell_text(1, 1, "no")
    grid.wait_for_cell_text(2, 0, "large")
    grid.wait_for_cell_text(2, 1, "yes")



from dash import Dash, html, Output, Input, no_update, State, ctx, Patch, dcc
import dash_ag_grid as dag
import plotly.express as px
import json
import time
import pandas as pd

from . import utils
from dash.testing.wait import until


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]

columnDefs = [
        {"headerName": "Make", "field": "make"},
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

alt_columnDefs = [
    {"field": "price", "pinned": False, "sort": "asc"},
    {"field": "model", "pinned": False},
    {"field": "make", "pinned": False},
]

defaultColDef = {
    "initialWidth": 150,
    "sortable": True,
    "resizable": True,
    "filter": True,
}

rowData = [
    {"make": "Toyota", "model": "Celica", "price": 35000},
    {"make": "Ford", "model": "Mondeo", "price": 32000},
    {"make": "Porsche", "model": "Boxster", "price": 72000},
]

colState = [
    {
        "colId": "make",
        "width": 150,
        "hide": False,
        "pinned": "left",
        "sort": None,
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
    {
        "colId": "price",
        "width": 150,
        "hide": False,
        "pinned": "left",
        "sort": None,
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
    {
        "colId": "model",
        "width": 150,
        "hide": False,
        "pinned": None,
        "sort": None,
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
]

alt_colState = [
    {
        "colId": "price",
        "width": 200,
        "hide": False,
        "pinned": None,
        "sort": "asc",
        "sortType": "default",
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
    {
        "colId": "model",
        "width": 150,
        "hide": False,
        "pinned": None,
        "sort": None,
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
    {
        "colId": "make",
        "width": 150,
        "hide": False,
        "pinned": None,
        "sort": None,
        "sortIndex": None,
        "aggFunc": None,
        "rowGroup": False,
        "rowGroupIndex": None,
        "pivot": False,
        "pivotIndex": None,
        "flex": None,
    },
]


def test_cs001_column_state(dash_duo):
    app = Dash(__name__)


    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Button(
                        "Reset Column State", id="reset-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Get Column State", id="get-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Load State", id="load-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Change Column Defs", id="load-column-defs", n_clicks=0
                    ),
                    html.Button(
                        "Load Both Column Defs and State",
                        id="load-column-state-defs-button",
                        n_clicks=0,
                    ),
                ],
            ),
            dag.AgGrid(
                id="grid",
                # columnSize="autoSize",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowData=rowData,
                columnState=colState,
            ),
            html.Div(id="reset-column-state-grid-pre"),
        ]
    )

    @app.callback(
        Output("grid", "resetColumnState"),
        Output("grid", "updateColumnState"),
        Input("reset-column-state-button", "n_clicks"),
        Input("get-column-state-button", "n_clicks"),
    )
    def reset_column_state(n_reset, n_state):
        if ctx.triggered_id == "reset-column-state-button":
            return True, False
        elif ctx.triggered_id == "get-column-state-button":
            return False, True
        return no_update

    @app.callback(
        Output("grid", "columnState"),
        Input("load-column-state-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return colState
        return no_update

    @app.callback(
        Output("reset-column-state-grid-pre", "children"),
        Input("grid", "columnState"),
    )
    def display_column_state(col_state):
        return json.dumps(col_state)

    @app.callback(
        Output("grid", "columnDefs"),
        Input("load-column-defs", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return alt_columnDefs
        return no_update

    @app.callback(
        Output("grid", "columnState", allow_duplicate=True),
        Output("grid", "columnDefs", allow_duplicate=True),
        Input("load-column-state-defs-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return colState, columnDefs
        return no_update, no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    grid.wait_for_pinned_cols(2)
    grid.wait_for_viewport_cols(1)

    dash_duo.find_element("#get-column-state-button").click()
    time.sleep(0.5)  # pausing to emulate separation because user inputs

    until(
        lambda: json.dumps(colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )

    grid.resize_col(1, 50)

    dash_duo.find_element("#get-column-state-button").click()
    testState = colState.copy()
    testState[1]["width"] = 200
    until(
        lambda: json.dumps(testState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )

    dash_duo.find_element("#load-column-defs").click()

    time.sleep(0.5)  # pausing to emulate separation because user inputs

    until(
        lambda: json.dumps(alt_colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )
    grid.wait_for_all_header_texts(["Price", "Model", "Make"])
    grid.wait_for_cell_text(0, 0, "32000")

    time.sleep(0.5)  # pausing to emulate separation because user inputs

    dash_duo.find_element("#load-column-state-defs-button").click()
    until(
        lambda: json.dumps(colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )
    grid.wait_for_all_header_texts(["Make", "Price", "Model"])

def test_cs002_column_state(dash_duo):
    app = Dash(__name__)

    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Button(
                        "Add Grid", id="add-grid", n_clicks=0
                    ),
                    html.Div(
                        id='grid-holder'
                    )
                ],
            ),
        ]
    )

    app.clientside_callback(
        """async ()=> {
            await new Promise(resolve => setTimeout(resolve, 400));
            return []
        }""",
        Output('grid-holder', 'children', allow_duplicate=True),
        Input('grid-holder', 'children'),
        prevent_initial_call=True
    )

    @app.callback(
        Output('grid-holder', 'children'),
        Input('add-grid', 'n_clicks'),
        prevent_initial_call=True
    )
    def make_grid(n):
        return dag.AgGrid(
                id=f"grid_{n}",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowData=rowData,
                columnSize='Auto'
            )

    dash_duo.start_server(app)

    for x in range(10):
        dash_duo.find_element("#add-grid").click()
        time.sleep(2)  # pausing to emulate separation because user inputs
    assert list(filter(lambda i: i.get("level") != "WARNING", dash_duo.get_logs())) == []

def test_cs003_column_state(dash_duo):
    data = [
        {
            "localTime": "5:00am",
            "a": 0.231,
            "b": 0.523,
            "c": 0.423,
            "d": 0.527,
        },
        {
            "localTime": "5:15am",
            "a": 0.423,
            "b": 0.452,
            "c": 0.523,
            "d": 0.543,
        },
        {
            "localTime": "5:30am",
            "a": 0.537,
            "b": 0.246,
            "c": 0.426,
            "d": 0.421,
        },
        {
            "localTime": "5:45am",
            "a": 0.893,
            "b": 0.083,
            "c": 0.532,
            "d": 0.983,
        },
        {
            "localTime": "6:00am",
            "a": 0.231,
            "b": 0.523,
            "c": 0.423,
            "d": 0.527,
        },
        {
            "localTime": "6:15am",
            "a": 0.423,
            "b": 0.452,
            "c": 0.523,
            "d": 0.543,
        },
        {
            "localTime": "6:30am",
            "a": 0.537,
            "b": 0.246,
            "c": 0.426,
            "d": 0.421,
        },
        {
            "localTime": "6:45am",
            "a": 0.893,
            "b": 0.083,
            "c": 0.532,
            "d": 0.983,
        },
        {
            "localTime": "7:00am",
            "a": 0.231,
            "b": 0.523,
            "c": 0.423,
            "d": 0.527,
        },
        {
            "localTime": "7:15am",
            "a": 0.423,
            "b": 0.452,
            "c": 0.523,
            "d": 0.543,
        },
        {
            "localTime": "7:30am",
            "a": 0.537,
            "b": 0.246,
            "c": 0.426,
            "d": 0.421,
        },
        {
            "localTime": "7:45am",
            "a": 0.893,
            "b": 0.083,
            "c": 0.532,
            "d": 0.983,
        },
        {
            "localTime": "8:00am",
            "a": 0.231,
            "b": 0.523,
            "c": 0.423,
            "d": 0.527,
        },
        {
            "localTime": "8:15am",
            "a": 0.423,
            "b": 0.452,
            "c": 0.523,
            "d": 0.543,
        },
        {
            "localTime": "8:30am",
            "a": 0.537,
            "b": 0.246,
            "c": 0.426,
            "d": 0.421,
        },
        {
            "localTime": "8:45am",
            "a": 0.893,
            "b": 0.083,
            "c": 0.532,
            "d": 0.983,
        },
        {
            "localTime": "8:00am",
            "a": 0.231,
            "b": 0.523,
            "c": 0.423,
            "d": 0.527,
        },
        {
            "localTime": "8:15am",
            "a": 0.423,
            "b": 0.452,
            "c": 0.523,
            "d": 0.543,
        },
        {
            "localTime": "8:30am",
            "a": 0.537,
            "b": 0.246,
            "c": 0.426,
            "d": 0.421,
        },
        {
            "localTime": "8:45am",
            "a": 0.893,
            "b": 0.083,
            "c": 0.532,
            "d": 0.983,
        },
    ]

    columnDefs = [
        {"field": "localTime"},
        {"field": "a"},
        {"field": "b"},
        {"field": "c"},
        {"field": "d"},
    ]

    app = Dash(__name__)

    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Button("Remove Column", id="remove-column", n_clicks=0),
                    html.Div(
                        id="grid-holder",
                        children=[
                            dag.AgGrid(
                                id=f"grid",
                                columnDefs=columnDefs,
                                rowData=data,
                                columnSize="autoSize",
                            )
                        ],
                    ),
                ],
            ),
        ]
    )

    @app.callback(
        Output("grid", "columnDefs"),
        Input("remove-column", "n_clicks"),
    )
    def remove_column(n):
        if n:
            cols = Patch()
            if n < 3:
                del cols[0]
                return cols
        return no_update

    dash_duo.start_server(
        app,
        debug=True,
        use_reloader=False,
        use_debugger=True,
        dev_tools_hot_reload=False,
        dev_tools_props_check=True,
        dev_tools_disable_version_check=True,
    )

    for x in range(3):
        dash_duo.find_element("#remove-column").click()
        time.sleep(2)  # pausing to emulate separation because user inputs
    assert list(filter(lambda i: i.get("level") != "ERROR", dash_duo.get_logs())) == []

def test_toggle_column_visibility(dash_duo):
    data = pd.DataFrame([
        {"a": 1, "b": 2},
        {"a": 3, "b": 4},
    ])

    app = Dash(__name__)

    app.layout = html.Div([
        dcc.Dropdown(
            id="select-columns",
            value=list(data.columns),
            options=[{"label": col, "value": col} for col in data.columns],
            multi=True,
        ),
        dag.AgGrid(
            id="ag-grid",
            style={"height": "75vh", "width": "100%"},
            rowData=data.to_dict(orient="records"),
        ),
    ])

    @app.callback(
        Output("ag-grid", "columnDefs"),
        Input("select-columns", "value"),
    )
    def toggle_column_visibility(selected_columns):
        if not selected_columns:
            return no_update
        return [
            {
                "headerName": col_name,
                "field": col_name,
                "hide": col_name not in selected_columns,
            }
            for col_name in data.columns
        ]

    dash_duo.start_server(app)

    # Wait for grid to render
    grid = utils.Grid(dash_duo, "ag-grid")

    grid.wait_for_cell_text(0, 0, "1")

    # Hide column 'b'
    dropdown = dash_duo.find_element("#select-columns")
    dropdown.click()
    option_b = dash_duo.find_element('label:nth-child(2) > span.dash-options-list-option-wrapper > input')
    option_b.click()
    time.sleep(1)

    # Only column 'a' should be visible
    grid_headers = dash_duo.find_elements("div.ag-header-cell-label")
    header_texts = [h.text for h in grid_headers]
    assert "a" in header_texts
    assert "b" not in header_texts

    # Show both columns again
    option_b.click()
    time.sleep(1)
    grid_headers = dash_duo.find_elements("div.ag-header-cell-label")
    header_texts = [h.text for h in grid_headers]
    assert "a" in header_texts
    assert "b" in header_texts



import json
import os

def get_package_info():
    basepath = os.path.dirname(__file__)
    filepath = os.path.abspath(os.path.join(basepath, 'package-info.json'))
    with open(filepath) as f:
        return json.load(f)

package = get_package_info()
__version__ = package['version']
grid_version = package['dependencies']['ag-grid-community']


from dash import Dash, html
import dash_ag_grid as dag
import plotly.express as px
import pandas as pd

from . import utils

df = px.data.election()
default_display_cols = ["district_id"]
other_cols = ["district", "winner"]

df = pd.concat([df, pd.DataFrame({"district": ["test"]})])


def test_fi002_custom_filter(dash_duo):
    app = Dash(__name__)

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                rowData=df.to_dict("records"),
                columnDefs=[
                               {
                                   "headerName": col.capitalize(),
                                   "field": col,
                                   "filterParams": {"function": "filterParams()"},
                                   "filter": "agNumberColumnFilter",
                               }
                               for col in default_display_cols
                           ] + [
                               {
                                   "headerName": col.capitalize(),
                                   "field": col,
                                   "filterParams": {
                                       "filterOptions": ["contains", "startsWith", "endsWith"],
                                       "defaultOption": "endsWith",
                                   },
                                   "filter": True,
                               }
                               for col in other_cols
                           ],
                defaultColDef={"floatingFilter": True},
            )
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 1, "101-Bois-de-Liesse")

    grid.set_filter(0, "12")

    grid.wait_for_cell_text(0, 1, "11-Sault-au-Récollet")
    grid.wait_for_rendered_rows(2)

    grid.set_filter(0, "")
    grid.wait_for_cell_text(0, 1, "101-Bois-de-Liesse")

    grid.set_filter(1, "t")
    grid.set_filter(2, "e")
    grid.wait_for_cell_text(0, 1, "11-Sault-au-Récollet")
    grid.wait_for_rendered_rows(8)


def test_fi003_custom_filter(dash_duo):
    app = Dash(__name__)

    df = pd.read_json('https://www.ag-grid.com/example-assets/olympic-winners.json', convert_dates=False)

    rowData = df.to_dict('records')

    columnDefs = [
        {'field': 'age', 'filter': 'agNumberColumnFilter'},
        {'field': 'country', 'minWidth': 150},
        {'field': 'year', 'filter': {'function': 'YearFilter'}},
        {
            'field': 'date',
            'minWidth': 130,
            'filter': 'agDateColumnFilter',
        },
        {'field': 'sport'},
        {'field': 'gold', 'filter': 'agNumberColumnFilter'},
        {'field': 'silver', 'filter': 'agNumberColumnFilter'},
        {'field': 'bronze', 'filter': 'agNumberColumnFilter'},
        {'field': 'total', 'filter': 'agNumberColumnFilter'},
    ]

    defaultColDef = {
        'editable': True,
        'sortable': True,
        'flex': 1,
        'minWidth': 100,
        'filter': True,
        'resizable': True,
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef=defaultColDef
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "23")

    dash_duo.find_element('.ag-header-cell[aria-colindex="3"] span[data-ref="eFilterButton"]').click()

    dash_duo.find_element('.ag-filter label:nth-child(2)').click()

    grid.wait_for_cell_text(0, 0, "27")

    dash_duo.find_element('.ag-filter label:nth-child(1)').click()

    grid.wait_for_cell_text(0, 0, "23")


def test_fi003_custom_filter_v34(dash_duo):
    app = Dash(__name__)

    df = pd.read_json('https://www.ag-grid.com/example-assets/olympic-winners.json', convert_dates=False)

    rowData = df.to_dict('records')

    columnDefs = [
        {'field': 'age', 'filter': 'agNumberColumnFilter'},
        {'field': 'country', 'minWidth': 150},
        {'field': 'year', 'filter': {'component': 'YearFilter34', 'doesFilterPass': {'function': 'doesFilterPass'}}},
        {
            'field': 'date',
            'minWidth': 130,
            'filter': 'agDateColumnFilter',
        },
        {'field': 'sport'},
        {'field': 'gold', 'filter': 'agNumberColumnFilter'},
        {'field': 'silver', 'filter': 'agNumberColumnFilter'},
        {'field': 'bronze', 'filter': 'agNumberColumnFilter'},
        {'field': 'total', 'filter': 'agNumberColumnFilter'},
    ]

    defaultColDef = {
        'editable': True,
        'sortable': True,
        'flex': 1,
        'minWidth': 100,
        'filter': True,
        'resizable': True,
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef=defaultColDef,
                dashGridOptions={"enableFilterHandlers": True}
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "23")

    dash_duo.find_element('.ag-header-cell[aria-colindex="3"] span[data-ref="eFilterButton"]').click()

    dash_duo.find_element('.ag-filter label:nth-child(2)').click()

    grid.wait_for_cell_text(0, 0, "27")

    dash_duo.find_element('.ag-filter label:nth-child(1)').click()

    grid.wait_for_cell_text(0, 0, "23")


# test filterParams.textFormatter, filterParams.textMatcher and filterParams.filterOptions.predicate functions
def test_fi004_custom_filter(dash_duo):
    app = Dash(__name__)

    df = pd.read_json('https://www.ag-grid.com/example-assets/olympic-winners.json', convert_dates=False)

    columnDefs = [
        {
            "field": "athlete",
            "filterParams": {"textFormatter": {"function": "myTextFormatter(params)"}},
        },
        {
            "field": "country",
            "filterParams": {"textMatcher": {"function": "myTextMatcher(params)"}},
        },
        {
            "field": "athlete",
            "filterParams": {
                "filterOptions": [
                    {
                        "displayKey": 'nameStartsWith',
                        "displayName": 'Name starts with',
                        "predicate": {"function": "startWith"},
                        "numberOfInputs": 1,
                    },
                ],
            },
        },
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                rowData=df.to_dict("records"),
                columnDefs=columnDefs,
                columnSize="sizeToFit",
                defaultColDef={"filter": True, "floatingFilter": True}
            ),
        ]
    )

    dash_duo.start_server(app)
    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    # Test textFormatter
    grid.set_filter(0, "bjo")
    grid.wait_for_cell_text(0, 0, "Björn Lind")

    # Remove filter
    grid.set_filter(0, "")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    # Test textMatcher
    grid.set_filter(1, "sean")
    grid.wait_for_cell_text(0, 1, "South Africa")

    # Remove filter
    grid.set_filter(1, "")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    # Test filterOptions
    grid.set_filter(2, "c")
    grid.wait_for_cell_text(0, 2, "Natalie Coughlin")


# test numberParser and numberFormatter functions in filterParams
def test_fi005_custom_filter(dash_duo):
    app = Dash(__name__)

    rowData = [{"sale": (i - 3) * 100} for i in range(50)]

    columnDefs = [
        {
            "field": "sale",
            "headerName": "Sale",
            "filter": "agNumberColumnFilter",
            "filterParams": {
                "filterOptions": ["greaterThan"],
                "allowedCharPattern": "\\d\\-\\,\\$",
                "numberParser": {"function": "myNumberParser(params)"},
                "numberFormatter": {"function": "myNumberFormatter(params)"},
            },
            "valueFormatter": {
                "function": "d3.formatLocale({'decimal': ',', 'thousands': '.', 'currency': ['$', '']}).format('$,.2f')(params.value)"
            },
        },
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                defaultColDef={"floatingFilter": True},
                rowData=rowData,
                columnSize="sizeToFit",
            ),
        ]
    )

    dash_duo.start_server(app)
    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "−$300,00")

    # Test numberParser and numberFormatter
    grid.set_filter(0, "$100,5")
    grid.wait_for_cell_text(0, 0, "$200,00")

def test_fi006_custom_filter(dash_duo):
    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    columnDefs = [
        {"field": "athlete",
        "filter": "agMultiColumnFilter",
                "filterParams": {
                    "filters": [
                        {"filter": "agTextColumnFilter"},
                        {"filter": "agSetColumnFilter"} # Example with Set Filter
                    ]
                }},
        {"field": "country"},
        {
            "field": "date",
            "filter": "agMultiColumnFilter",
            "filterParams": {
                "filters": [
                    {
                        "filter": "agSetColumnFilter",
                        'filterParams': {'excelMode': 'windows', 'buttons': ['apply', 'reset'],
                        }
                    },
                    {
                        "filter": "agDateColumnFilter",
                        'filterParams': {
                            'excelMode': 'windows',
                            'buttons': ['apply', 'reset'],
                            'comparator': {'function': 'dateFilterComparator'},
                        }
                    },
                ],

            },
        },
    ]


    app.layout = html.Div(
        [
            dag.AgGrid(
                id="date-filter-example",
                enableEnterpriseModules=True,
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                defaultColDef={"flex": 1, "minWidth": 150, "floatingFilter": True},
                dashGridOptions={"animateRows": False}
            ),
        ],
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "date-filter-example")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    # Test Set Filter - click filter button on date column
    dash_duo.find_element('.ag-floating-filter[aria-colindex="3"] button').click()

    # Uncheck "Select All"
    dash_duo.find_element('.ag-set-filter-list .ag-set-filter-item .ag-checkbox-input').click()

    # Select "24/08/2008"
    dash_duo.wait_for_element('.ag-set-filter-list .ag-virtual-list-item', timeout=10)
    set_filter_items = dash_duo.find_elements('.ag-set-filter-list .ag-virtual-list-item')
    checkboxes = dash_duo.find_elements('.ag-set-filter-list .ag-virtual-list-item .ag-checkbox-input')

    for i, item in enumerate(set_filter_items):
        if "24/08/2008" in item.text:
            checkboxes[i].click()
            break

    # Apply
    dash_duo.find_element('button[data-ref="applyFilterButton"]').click()
    grid.wait_for_cell_text(0, 2, "24/08/2008")

    # Reset
    dash_duo.find_element('.ag-floating-filter[aria-colindex="3"] button').click()
    dash_duo.find_element('button[data-ref="resetFilterButton"]').click()

    # Test Date Filter - click filter button again
    dash_duo.find_element('.ag-floating-filter[aria-colindex="3"] button').click()

    # Type date
    date_input = dash_duo.find_element('.ag-filter-wrapper .ag-date-filter input[class="ag-input-field-input ag-text-field-input"]')
    date_input.click()
    date_input.send_keys("24-08-2008")

    # Apply
    apply_buttons = dash_duo.find_elements('button[data-ref="applyFilterButton"]')
    apply_buttons[1].click()
    grid.wait_for_cell_text(0, 2, "24/08/2008")